﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Grafo_Consola
{
    class ClaseGrafoNoPonderadoDirigido
    {
        // Lista simple que almacena los nodos del grafo
        ClaseListaSimpleOrdenada<ClaseNodoGrafo> ListaOrdenadaNodos;

        // Constructor
        public ClaseGrafoNoPonderadoDirigido()
        {
            // Crea e inicializa vacía la lista simple de nodos ordenados
            ListaOrdenadaNodos = new ClaseListaSimpleOrdenada<ClaseNodoGrafo>();
        }

        // Método público para detectar si el grafo está vacío
        public bool EstaVacio()
        {
            if (ListaOrdenadaNodos.EstaVacia())
                return (true);
            else
                return (false);
        }

        // Método público para insertar un nodo en el grafo
        public void InsertarNodo(int Dato)
        {
            ClaseNodoGrafo Nodo = new ClaseNodoGrafo();
            Nodo.Dato = Dato;

            // Inserta el Nodo en la Lista de nodos
            ListaOrdenadaNodos.Insertar(Nodo);
        }

        // Método público para eliminar un nodo y sus arcos del grafo
        public void EliminarNodo(int Dato)
        {
            ClaseNodoGrafo Nodo = new ClaseNodoGrafo();

            Nodo.Dato = Dato;

            // Busca el nodo en la lista de nodos
            Nodo = ListaOrdenadaNodos.Buscar(Nodo);
            if (Nodo == null)
                throw new Exception("No está registrado ...");

            // Crea una lista temporal de los arcos del nodo que se desea eliminar
            ClaseListaSimpleOrdenada<ClaseArcos> ListaArcos = new ClaseListaSimpleOrdenada<ClaseArcos>();
            foreach (ClaseArcos Arco in Nodo.ListaOrdenadaArcos)
                ListaArcos.Insertar(Arco);

            // Crea una lista temporal de los arcos inversos del nodo que se desea eliminar
            ClaseListaSimpleOrdenada<ClaseArcosInversos> ListaArcosInversos = new ClaseListaSimpleOrdenada<ClaseArcosInversos>();
            foreach (ClaseArcosInversos ArcoInverso in Nodo.ListaOrdenadaArcosInversos)
                ListaArcosInversos.Insertar(ArcoInverso);

            // Elimina cada arco
            //foreach (ClaseArcos Arco in ListaArcos)
            //    Nodo.ListaOrdenadaArcos.Eliminar(Arco);

            foreach (ClaseArcos Arco in ListaArcos)
                Arco.Destino.ListaOrdenadaArcosInversos.Eliminar(new ClaseArcosInversos() { Origen = Nodo });

            // Elimina cada arco inverso
            foreach (ClaseArcosInversos arcosInverso in ListaArcosInversos)
            {
                arcosInverso.Origen.ListaOrdenadaArcos.Eliminar(new ClaseArcos() { Destino = Nodo });
            }

            //foreach (ClaseArcosInversos ArcoInverso in ListaArcosInversos)
            //    Nodo.ListaOrdenadaArcosInversos.Eliminar(ArcoInverso);

            // Elimina el nodo
            ListaOrdenadaNodos.Eliminar(Nodo);
        }

        // Método público para buscar y devolver un nodo del grafo
        public ClaseNodoGrafo BuscarNodo(int Dato)
        {
            ClaseNodoGrafo Nodo = new ClaseNodoGrafo();
            Nodo.Dato = Dato;

            // Busca el nodo en la lista de nodos
            return (ListaOrdenadaNodos.Buscar(Nodo));
        }

        // Método público para insertar un arco y su respectivo arco inverso en el grafo
        public void InsertarArco(int DatoOrigen, int DatoDestino)
        {
            ClaseNodoGrafo NodoOrigen = new ClaseNodoGrafo();
            ClaseNodoGrafo NodoDestino = new ClaseNodoGrafo();

            NodoOrigen.Dato = DatoOrigen;
            NodoOrigen = ListaOrdenadaNodos.Buscar(NodoOrigen);  // Busca el nodo origen

            NodoDestino.Dato = DatoDestino;
            NodoDestino = ListaOrdenadaNodos.Buscar(NodoDestino);  // Busca el nodo destino

            // Inserta el arco en la lista de arcos del nodo origen
            ClaseArcos Arco = new ClaseArcos();
            Arco.Destino = ListaOrdenadaNodos.Buscar(new ClaseNodoGrafo() { Dato = DatoDestino });
            NodoOrigen.ListaOrdenadaArcos.Insertar(Arco);

            // Inserta el arco inverso en la lista de arcos inversos del nodo destino
            ClaseArcosInversos ArcoInverso = new ClaseArcosInversos();
            ArcoInverso.Origen = ListaOrdenadaNodos.Buscar(new ClaseNodoGrafo() { Dato = DatoOrigen });
            NodoDestino.ListaOrdenadaArcosInversos.Insertar(ArcoInverso);

            return;
        }

        // Método público para eliminar un arco y su respectivo arco inverso
        public void EliminarArco(int DatoOrigen, int DatoDestino)
        {
            ClaseNodoGrafo NodoOrigen = new ClaseNodoGrafo();
            ClaseNodoGrafo NodoDestino = new ClaseNodoGrafo();

            NodoOrigen.Dato = DatoOrigen;
            NodoOrigen = ListaOrdenadaNodos.Buscar(NodoOrigen);  // Busca el nodo origen

            NodoDestino.Dato = DatoDestino;
            NodoDestino = ListaOrdenadaNodos.Buscar(NodoDestino);  // Busca el nodo destino

            // Elimina el arco de la lista de arcos del nodo origen
            ClaseArcos Arco = new ClaseArcos();
            Arco.Destino = ListaOrdenadaNodos.Buscar(new ClaseNodoGrafo() { Dato = DatoDestino });
            NodoOrigen.ListaOrdenadaArcos.Eliminar(Arco);

            // Elimina el arco inverso de la lista de arcos inversos del nodo destino
            ClaseArcosInversos ArcoInverso = new ClaseArcosInversos();
            ArcoInverso.Origen = ListaOrdenadaNodos.Buscar(new ClaseNodoGrafo() { Dato = DatoOrigen });
            NodoDestino.ListaOrdenadaArcosInversos.Eliminar(ArcoInverso);

            return;
        }

        // Método público que devuelve una cadena con la información de nodos, arcos y
        // arcos inversos de un grafo
        public string Mostrar()
        {
            string Resultado = "";

            if (ListaOrdenadaNodos.EstaVacia())
                Resultado = Resultado + "\nGrafo vacío ...";
            else
            {
                // Recorre la lista de nodos del grafo para mostrar sus arcos
                Resultado = Resultado + "\nNODO       ARCOS";
                foreach (ClaseNodoGrafo NodoActual in ListaOrdenadaNodos)
                {
                    Resultado = Resultado + "\n[" + NodoActual.Dato.ToString() + "] -> ";

                    if (!NodoActual.ListaOrdenadaArcos.EstaVacia())
                    {
                        // Recorre la lista de arcos de cada nodo
                        foreach (ClaseArcos Arco in NodoActual.ListaOrdenadaArcos)
                            Resultado = Resultado + "(" + NodoActual.Dato + "," + Arco.Destino + ") ";
                    }
                }


                // Recorre la lista de nodos del grafo para mostrar sus arcos inversos
                Resultado = Resultado + "\n\nNODO       ARCOS INVERSOS";
                foreach (ClaseNodoGrafo NodoActual in ListaOrdenadaNodos)
                {
                    Resultado = Resultado + "\n[" + NodoActual.Dato.ToString() + "] <- ";

                    if (!NodoActual.ListaOrdenadaArcosInversos.EstaVacia())
                    {
                        // Recorre la lista de arcos inversos de cada nodo
                        foreach (ClaseArcosInversos ArcoInverso in NodoActual.ListaOrdenadaArcosInversos)
                            Resultado = Resultado + "(" + ArcoInverso.Origen + "," + NodoActual.Dato + ") ";
                    }
                }
            }
            return (Resultado);
        }

        // Método público que recorre en anchura el grafo y devuelve una cadena 
        // con la secuencia de nodos visitados
        public string RecorridoAnchura(int DatoOrigen, int DatoDestino)
        {
            string Resultado = "";
            // Crea la estructura de nodos en espera (cola)
            ClaseColaLista<ClaseNodoCamino> ColaDesordenadaNodosEnEspera = new ClaseColaLista<ClaseNodoCamino>();

            // Crea la estructura de nodos examinados (cola)
            ClaseListaDobleDesordenada<ClaseNodoCamino> ListaDobleDesordenadaNodosExaminados = new ClaseListaDobleDesordenada<ClaseNodoCamino>();

            ClaseNodoGrafo NodoGrafo;
            ClaseNodoCamino Descendiente;
            ClaseListaDobleDesordenada<ClaseNodoCamino> ListaDobleDesordenadaTrayectoria;

            string Nodo, Antecesor;
            ClaseNodoCamino NodoCamino = new ClaseNodoCamino();
            NodoCamino.Dato = DatoOrigen.ToString();
            NodoCamino.Origen = "";

            ColaDesordenadaNodosEnEspera.Insertar(NodoCamino);


            // Mientras haya nodos en espera ...
            while (!ColaDesordenadaNodosEnEspera.EstaVacia())
            {
                // Saca y elimina el primer nodo en espera
                NodoCamino = ColaDesordenadaNodosEnEspera.Eliminar();

                // Si es el nodo destino ...
                if (NodoCamino.Dato == DatoDestino.ToString())
                {
                    // Inserta en nodo en la estructura de nodos examinados
                    ListaDobleDesordenadaNodosExaminados.Insertar(NodoCamino);

                    ListaDobleDesordenadaTrayectoria = new ClaseListaDobleDesordenada<ClaseNodoCamino>();
                    Antecesor = DatoDestino.ToString();
                    Nodo = "";

                    // Recorre la estructura de nodos examinados para mostrar la trayectoria
                    foreach (ClaseNodoCamino NodoEnTrayectoria in ListaDobleDesordenadaNodosExaminados.DesdeNodoFinal)
                    {
                        if (NodoEnTrayectoria.Dato == Antecesor)
                        {
                            ListaDobleDesordenadaTrayectoria.Insertar(NodoEnTrayectoria);
                            Antecesor = NodoEnTrayectoria.Origen;
                            Nodo = NodoEnTrayectoria.Dato;
                        }
                    }

                    foreach (ClaseNodoCamino NodoEnTrayectoria in ListaDobleDesordenadaTrayectoria.DesdeNodoFinal)
                        Resultado = Resultado + "-> " + NodoEnTrayectoria.Dato;

                    // Devuelve la trayectoria
                    return (Resultado);
                }

                ListaDobleDesordenadaNodosExaminados.Insertar(NodoCamino);

                NodoGrafo = new ClaseNodoGrafo();
                NodoGrafo.Dato = int.Parse(NodoCamino.Dato);
                NodoGrafo = ListaOrdenadaNodos.Buscar(NodoGrafo);

                // Recorre la lista de arcos del nodo
                foreach (ClaseArcos Arco in NodoGrafo.ListaOrdenadaArcos)
                {
                    // Crea un descendiente del nodo
                    Descendiente = new ClaseNodoCamino();

                    Descendiente.Dato = Arco.Destino.ToString();
                    Descendiente.Origen = NodoGrafo.Dato.ToString();

                    // Si el descendiente no está en espera ni ha sido examinado previamente ...
                    if (ColaDesordenadaNodosEnEspera.Buscar(Descendiente) == null && ListaDobleDesordenadaNodosExaminados.Buscar(Descendiente) == null)
                        ColaDesordenadaNodosEnEspera.Insertar(Descendiente);
                }
            }
            // Termina el ciclo y no encuentra el camino
            return ("No hay camino entre " + DatoOrigen.ToString() + " y " + DatoDestino.ToString());
        }

        public string RecorridoProfundidad(int DatoOrigen, int DatoDestino)
        {
            string Resultado = "";
            // Crea la estructura de nodos en espera (cola)
            ClasePilaDesordenadaLista<ClaseNodoCamino> PilaDesordenadaNodosEnEspera = new ClasePilaDesordenadaLista<ClaseNodoCamino>();

            // Crea la estructura de nodos examinados (cola)
            ClaseListaDobleDesordenada<ClaseNodoCamino> ListaDobleDesordenadaNodosExaminados = new ClaseListaDobleDesordenada<ClaseNodoCamino>();

            ClaseNodoGrafo NodoGrafo;
            ClaseNodoCamino Descendiente;
            ClaseListaDobleDesordenada<ClaseNodoCamino> ListaDobleDesordenadaTrayectoria;

            string Nodo, Antecesor;
            ClaseNodoCamino NodoCamino = new ClaseNodoCamino();
            NodoCamino.Dato = DatoOrigen.ToString();
            NodoCamino.Origen = "";

            PilaDesordenadaNodosEnEspera.Push(NodoCamino);


            // Mientras haya nodos en espera ...
            while (!PilaDesordenadaNodosEnEspera.EstaVacia())
            {
                // Saca y elimina el primer nodo en espera
                NodoCamino = PilaDesordenadaNodosEnEspera.Pop();

                // Si es el nodo destino ...
                if (NodoCamino.Dato == DatoDestino.ToString())
                {
                    // Inserta en nodo en la estructura de nodos examinados
                    ListaDobleDesordenadaNodosExaminados.Insertar(NodoCamino);

                    ListaDobleDesordenadaTrayectoria = new ClaseListaDobleDesordenada<ClaseNodoCamino>();
                    Antecesor = DatoDestino.ToString();
                    Nodo = "";

                    // Recorre la estructura de nodos examinados para mostrar la trayectoria
                    foreach (ClaseNodoCamino NodoEnTrayectoria in ListaDobleDesordenadaNodosExaminados.DesdeNodoFinal)
                    {
                        if (NodoEnTrayectoria.Dato == Antecesor)
                        {
                            ListaDobleDesordenadaTrayectoria.Insertar(NodoEnTrayectoria);
                            Antecesor = NodoEnTrayectoria.Origen;
                            Nodo = NodoEnTrayectoria.Dato;
                        }
                    }

                    foreach (ClaseNodoCamino NodoEnTrayectoria in ListaDobleDesordenadaTrayectoria.DesdeNodoFinal)
                        Resultado = Resultado + "-> " + NodoEnTrayectoria.Dato;

                    // Devuelve la trayectoria
                    return (Resultado);
                }

                ListaDobleDesordenadaNodosExaminados.Insertar(NodoCamino);

                NodoGrafo = new ClaseNodoGrafo();
                NodoGrafo.Dato = int.Parse(NodoCamino.Dato);
                NodoGrafo = ListaOrdenadaNodos.Buscar(NodoGrafo);

                // Recorre la lista de arcos del nodo
                foreach (ClaseArcos Arco in NodoGrafo.ListaOrdenadaArcos)
                {
                    // Crea un descendiente del nodo
                    Descendiente = new ClaseNodoCamino();

                    Descendiente.Dato = Arco.Destino.ToString();
                    Descendiente.Origen = NodoGrafo.Dato.ToString();

                    // Si el descendiente no está en espera ni ha sido examinado previamente ...
                    if (PilaDesordenadaNodosEnEspera.Buscar(Descendiente) == null && ListaDobleDesordenadaNodosExaminados.Buscar(Descendiente) == null)
                        PilaDesordenadaNodosEnEspera.Push(Descendiente);
                }
            }
            // Termina el ciclo y no encuentra el camino
            return ("No hay camino entre " + DatoOrigen.ToString() + " y " + DatoDestino.ToString());
        }
    }
}
